from .utils import GenomeFile, create_replace_function, split_locus_tag


class NoLocusTagInGffLine(KeyError):
    pass


class GffFile(GenomeFile):
    def rename(
            self,
            out: str,
            new_locus_tag_prefix: str,
            old_locus_tag_prefix: str = None,
            validate: bool = False
    ) -> None:
        old_locus_tag_prefix = self._pre_rename_check(out, new_locus_tag_prefix, old_locus_tag_prefix)

        with open(self.path) as in_f:
            content = in_f.read()

        replace_fn = create_replace_function({
            string.format(prefix=old_locus_tag_prefix): string.format(prefix=new_locus_tag_prefix)
            for string in ['-{prefix}', '={prefix}', ':{prefix}']
        })

        old_hash = hash(content)
        content = replace_fn(content)
        assert old_hash != hash(content), f'The content of {self.path=} has not changed!'

        assert new_locus_tag_prefix in content, f'Something went wrong: did not replace anything!'

        with open(out, 'w') as out_f:
            out_f.write(content)

        self.path = out

        if validate:
            self.validate_locus_tags(locus_tag_prefix=new_locus_tag_prefix)

    def detect_locus_tag_prefix(self) -> str:
        with open(self.path) as f:
            for line in f:
                if line.startswith('#'):
                    continue
                try:
                    locus_tag_prefix, gene_id = self._extract_gff_locus_tag(line)
                except NoLocusTagInGffLine:
                    continue
                return locus_tag_prefix

        raise KeyError(f'Could not extract locus_tag from {self.path=}')

    def validate_locus_tags(self, locus_tag_prefix: str = None):
        with open(self.path) as f:
            for line in f:
                if line == '##FASTA\n':
                    break  # prokka
                if line.startswith('#'):
                    continue
                try:
                    real_locus_tag_prefix, gene_id = self._extract_gff_locus_tag(line)
                except NoLocusTagInGffLine:
                    continue  # in PGAP gffs, some lines contain no locus_tag
                assert real_locus_tag_prefix == locus_tag_prefix, \
                    f'locus_tag_prefix in {self.path=} does not match. expected: {locus_tag_prefix} reality: {real_locus_tag_prefix}'

    @staticmethod
    def _extract_gff_data(line: str) -> dict:
        line = line.rstrip('\n').split('\t')
        assert len(line) == 9, f'gff line is malformed! {len(line)=} {line=}'
        return dict(info.split('=', 1) for info in line[8].split(';') if '=' in info)

    @classmethod
    def _extract_gff_locus_tag(cls, line: str) -> (str, str):
        data = cls._extract_gff_data(line)
        if 'locus_tag' not in data:
            raise NoLocusTagInGffLine(f'gff data contains no locus_tag! {line}')
        locus_tag = data['locus_tag']
        locus_tag_prefix, gene_id = split_locus_tag(locus_tag)
        assert ' ' not in locus_tag_prefix, f'The locus_tag may not contain blanks! {locus_tag=}'
        return locus_tag_prefix, gene_id


def rename_gff(
        file: str, out: str,
        new_locus_tag_prefix: str,
        old_locus_tag_prefix: str = None,
        validate: bool = False
):
    """
    Change the locus tags in a general feature format (gff) file

    :param file: input file
    :param out: output file
    :param new_locus_tag_prefix: desired locus tag
    :param old_locus_tag_prefix: locus tag to replace
    :param validate: if true, perform sanity check
    """
    GffFile(
        file=file
    ).rename(
        out=out,
        new_locus_tag_prefix=new_locus_tag_prefix,
        old_locus_tag_prefix=old_locus_tag_prefix,
        validate=validate
    )


def main():
    import fire

    fire.Fire(rename_gff)


if __name__ == '__main__':
    main()



import logging

from .utils import GenomeFile, split_locus_tag


class FastaFile(GenomeFile):
    def rename(self, out: str, new_locus_tag_prefix: str, old_locus_tag_prefix: str = None,
               validate: bool = False) -> None:
        old_locus_tag_prefix = self._pre_rename_check(out, new_locus_tag_prefix, old_locus_tag_prefix)

        with open(self.path) as in_f:
            content = in_f.readlines()

        def rename_line(line: str):
            if line.startswith('>'):
                assert old_locus_tag_prefix in line, \
                    f'Fasta header does not contain old_locus_tag_prefix! {old_locus_tag_prefix=}, {line=}, fasta={self.path}'
                return line \
                    .replace(old_locus_tag_prefix, new_locus_tag_prefix, 1) \
                    .replace(f'hypothetical protein {old_locus_tag_prefix}',
                             f'hypothetical protein {new_locus_tag_prefix}')
            else:
                return line

        content = [rename_line(line) for line in content]

        with open(out, 'w') as out_f:
            out_f.writelines(content)

        self.path = out

        if validate:
            self.validate_locus_tags(locus_tag_prefix=new_locus_tag_prefix)

    def detect_locus_tag_prefix(self) -> str:
        with open(self.path) as f:
            for line in f:
                if not line.startswith('>'):
                    assert line.strip() == '', f'Could not extract locus_tag from {self.path=}, it does not start with a header line!'
                    continue
                locus_tag_prefix, gene_id = self.parse_fasta_header(line)
                return locus_tag_prefix

        raise KeyError(
            f'Could not extract locus_tag from {self.path=}, it does not appear to contain a header line (>)!')

    def validate_locus_tags(self, locus_tag_prefix: str = None):
        with open(self.path) as f:
            for line in f:
                if line.startswith('>'):
                    real_locus_tag_prefix, gene_id = self.parse_fasta_header(header=line)
                assert real_locus_tag_prefix == locus_tag_prefix, \
                    f'locus_tag_prefix in {self.path=} does not match. expected: {locus_tag_prefix} reality: {real_locus_tag_prefix}'
                assert gene_id.isdigit(), f'locus_tag in {self.path=} is malformed. gene_id is expected to be: [0-9]+ reality: {gene_id}'

    @staticmethod
    def parse_fasta_header(header: str) -> (str, str):
        header = header.rstrip()
        error_message = f'This fasta file does not start with gene identifiers (>gene-identifier_00001)! {header=}'
        assert header.startswith('>'), error_message
        if not '_' in header:
            logging.warning(error_message)
        locus_tag = header[1:].split(' ', 1)[0]
        locus_tag_prefix, gene_id = split_locus_tag(locus_tag)
        assert ' ' not in locus_tag_prefix and len(locus_tag_prefix) > 0, error_message
        assert gene_id.isdigit(), error_message
        return locus_tag_prefix, gene_id


def rename_fasta(file: str, out: str, new_locus_tag_prefix: str, old_locus_tag_prefix: str = None,
                 validate: bool = False):
    """
    Change the locus tags in a protein/nucleotide FASTA file

    :param file: input file
    :param out: output file
    :param new_locus_tag_prefix: desired locus tag
    :param old_locus_tag_prefix: locus tag to replace
    :param validate: if true, perform sanity check
    """
    FastaFile(
        file=file
    ).rename(
        out=out,
        new_locus_tag_prefix=new_locus_tag_prefix,
        old_locus_tag_prefix=old_locus_tag_prefix,
        validate=validate
    )


def main():
    import fire

    fire.Fire(rename_fasta)


if __name__ == '__main__':
    main()



from schema import Schema, And, Optional, Or
from .utils import is_valid_date, get_cog_categories


class Dummy:
    def __repr__(self):
        return '<Dummy>'


dummy = Dummy()

organism_json_schema = Schema({
    'name': str,
    'alternative_name': Or(str, None),
    'taxid': int,
    'restricted': bool,
    'tags': [str],
    'representative': str
})

organism_json_dummy = {
    'name': dummy,
    'alternative_name': None,
    'taxid': dummy,
    'restricted': False,
    'tags': [],
    'representative': dummy
}

genome_json_schema = Schema({
    'identifier': And(str, lambda x: all(char not in x for char in ' _:/\\')),
    'contaminated': bool,
    'old_identifier': Or(str, None),
    'isolation_date': Or(is_valid_date, None),
    'env_broad_scale': [str],
    'env_local_scale': [str],
    'env_medium': [str],
    'growth_condition': Or(str, None),
    'geographical_coordinates': Or(str, None),
    'geographical_name': Or(str, None),
    'library_preparation': Or(str, None),
    'sequencing_tech': Or(str, None),
    'sequencing_tech_version': Or(str, None),
    'sequencing_date': Or(is_valid_date, None),
    'sequencing_coverage': Or(str, None),
    'read_length': Or(str, None),
    'assembly_tool': Or(str, None),
    'assembly_version': Or(str, None),
    'assembly_date': Or(is_valid_date, None),
    'nr_replicons': Or(int, None),
    'cds_tool': Or(str, None),
    'cds_tool_date': Or(is_valid_date, None),
    'cds_tool_version': Or(str, None),
    'cds_tool_faa_file': str,
    'cds_tool_ffn_file': str,
    'cds_tool_gbk_file': str,
    'cds_tool_gff_file': str,
    'cds_tool_sqn_file': Or(str, None),
    'assembly_fasta_file': str,
    'custom_annotations': [{
        "date": is_valid_date,
        "file": str,
        "type": str
    }],
    'BUSCO': Or({}, {
        'C': int,
        'D': int,
        'F': int,
        'M': int,
        'S': int,
        'T': int,
        Optional('dataset'): str,
        Optional('dataset_creation_date'): is_valid_date,
    }),
    'COG': {
        Optional(cog): Or(int, float)
        for cog in get_cog_categories().keys()
    },
    'bioproject_accession': Or(str, None),
    'biosample_accession': Or(str, None),
    'genome_accession': Or(str, None),
    'literature_references': [{
        "url": str,
        "name": str
    }],
    'custom_tables': dict,
    'tags': [str]
})

genome_json_dummy = {
    "identifier": dummy,
    "contaminated": False,
    "old_identifier": None,
    "isolation_date": None,
    "env_broad_scale": [],
    "env_local_scale": [],
    "env_medium": [],
    "growth_condition": None,
    "geographical_coordinates": None,
    "geographical_name": None,
    "library_preparation": None,
    "sequencing_tech": None,
    "sequencing_tech_version": None,
    "sequencing_date": None,
    "sequencing_coverage": None,
    "read_length": None,
    "assembly_tool": None,
    "assembly_version": None,
    "assembly_date": None,
    "nr_replicons": None,
    "cds_tool": None,
    "cds_tool_date": None,
    "cds_tool_version": None,
    "cds_tool_faa_file": dummy,
    "cds_tool_ffn_file": dummy,
    "cds_tool_gbk_file": dummy,
    "cds_tool_gff_file": dummy,
    "cds_tool_sqn_file": None,
    "assembly_fasta_file": dummy,
    "custom_annotations": [],
    "BUSCO": {},
    "COG": {},
    "bioproject_accession": None,
    "biosample_accession": None,
    "genome_accession": None,
    "literature_references": [],
    "custom_tables": {},
    "tags": []
}



__version__ = '0.0.6'
__folder_structure_version__ = 2

from .genbank_to_fasta import GenBankToFasta
from .import_genome import OgbImporter
from .metadata_schemas import organism_json_schema, genome_json_schema
from .reindex_assembly import reindex_assembly
from .rename_eggnog import rename_eggnog, EggnogFile
from .rename_gff import rename_gff, GffFile
from .rename_genbank import rename_genbank, GenBankFile
from .rename_custom_annotations import rename_custom_annotations, CustomAnnotationFile
from .rename_fasta import rename_fasta, FastaFile
from .download_ncbi_genome import download_ncbi_fna_gbk_gff, rename_ncbi_files, download_ncbi_genome



import os


def reindex_assembly(file: str, out: str, prefix: str, leading_zeroes: int = None):
    """
    Change the header line of FASTA to: f'>{prefix}_{counter}'.

    :param file: input file
    :param out: output file
    :param prefix: desired prefix
    :param leading_zeroes: format counter with leading zeroes (optional). e.g.: 5 -> >PREFIX_00001
    """
    assert not os.path.isfile(out), f'Output file already exists! {out=}'

    if type(leading_zeroes) is int and leading_zeroes > 1:
        format = lambda c: f'>{prefix}{str(c).zfill(leading_zeroes)}\n'
    else:
        format = lambda c: f'>{prefix}{c}\n'

    counter = 0
    with open(file) as in_f, open(out, 'w') as out_f:
        for line in in_f:
            if line.startswith('>'):
                counter += 1
                out_f.write(format(counter))
            else:
                out_f.write(line)


def main():
    import fire

    fire.Fire(reindex_assembly)


if __name__ == '__main__':
    main()



import gzip
import json
import logging
import os
import re
from datetime import datetime
from string import digits
from typing import Union, Callable

from Bio import Entrez
from termcolor import colored

DATE_FORMAT = '%Y-%m-%d'
TMPDIR = os.environ.get('TMPDIR', '/tmp')
Entrez.email = os.environ.get('ENTREZ_EMAIL', 'opengenomebrowser@bioinformatics.unibe.ch')

PACKAGE_ROOT = os.path.dirname(__file__)

ANNOTATIONS_JSON = f'{PACKAGE_ROOT}/data/annotations.json'
COG_CATEGORIES_JSON = f'{PACKAGE_ROOT}/data/COG_categories.json'
for f in [ANNOTATIONS_JSON, COG_CATEGORIES_JSON]:
    assert os.path.isfile(f), f'Package is poorly configured: file is missing: {f}'


class GenomeFile:
    original_path: str
    target_path: str

    def __init__(self, file: str, original_path: str = None):
        self.path = file
        self.original_path = original_path
        assert os.path.isfile(file), f'{self}: {file=} does not exist!'

    def __str__(self):
        return f'{type(self).__name__}: {os.path.basename(self.path)}'

    def metadata(self) -> (dict, dict):
        return {}, {}

    def detect_locus_tag_prefix(self) -> str:
        raise NotImplementedError('This function must be overwritten.')

    def validate_locus_tags(self, locus_tag_prefix: str = None):
        raise NotImplementedError('This function must be overwritten.')

    def rename(self, out: str, new_locus_tag_prefix: str, old_locus_tag_prefix: str = None, validate: bool = True) -> None:
        raise NotImplementedError('This function must be overwritten.')

    def _pre_rename_check(self, out: str, new_locus_tag_prefix: str, old_locus_tag_prefix: str = None) -> str:
        assert not os.path.isfile(out), f'Output file already exists! {out=}'

        if old_locus_tag_prefix is None:
            old_locus_tag_prefix = self.detect_locus_tag_prefix()
            logging.info(f'detected locus tag: {old_locus_tag_prefix}')
        else:
            detected_locus_tag_prefix = self.detect_locus_tag_prefix()
            logging.info(f'detected locus tag: {detected_locus_tag_prefix}')
            assert old_locus_tag_prefix == detected_locus_tag_prefix, f'Locus tag prefix does not match! {detected_locus_tag_prefix=} != {old_locus_tag_prefix=}'

        assert new_locus_tag_prefix != old_locus_tag_prefix, f'Aborting: {new_locus_tag_prefix=} is the same as {old_locus_tag_prefix=}'

        return old_locus_tag_prefix

    def date(self) -> datetime:
        return get_ctime(file=self.path)

    def date_str(self) -> str:
        """
        :returns date in this format: "%Y-%m-%d"
        """
        return date_to_string(self.date())


def query_yes_no(question: str, default: str = None, color='blue') -> bool:
    """
    Ask a yes/no question via raw_input() and return their answer

    :param question: string that is presented to the user
    :param default: the presumed answer if the user just hits <Enter>
    :param color: color in which the question is formatted
    :return: response (bool)
    """
    valid = {'yes': True, 'y': True, 'ye': True, 'no': False, 'n': False}
    if default is None:
        prompt = '[y/n] '
    elif default == 'yes':
        prompt = '[Y/n] '
    elif default == 'no':
        prompt = '[y/N] '
    else:
        raise ValueError(F"invalid default answer: '{default}'")

    while True:
        print(colored(F'{question} {prompt}', color), end='')
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            print(colored("'Please respond with 'y' or 'n'.'", color))


def query_int(question: str, error_msg='Please respond with an integer.', color='blue') -> int:
    while True:
        print(colored(question, color), end='')
        choice = input()
        if choice.isdigit():
            return int(choice)
        else:
            print(colored(error_msg, color))


def date_to_string(dt: datetime) -> str:
    return dt.strftime('%Y-%m-%d')


def get_ctime(file: str) -> datetime:
    ctime = os.path.getctime(file)
    dt = datetime.fromtimestamp(ctime)
    return dt


def is_valid_date(date: str) -> bool:
    try:
        datetime.strptime(date, DATE_FORMAT)
        return True
    except ValueError:
        return False


def get_cog_categories(reload: bool = False) -> dict:
    file = f'{PACKAGE_ROOT}/data/COG_categories.json'

    if not reload and os.path.isfile(file):
        with open(file) as f:
            cog_categories = json.load(f)
        return cog_categories

    else:
        from urllib import request
        with request.urlopen('https://ftp.ncbi.nih.gov/pub/COG/COG2020/data/fun-20.tab') as f:
            content = f.read().decode('utf-8').strip()

        def extract_cat(line: str) -> (str, dict):
            cat, color, description = line.split('\t')
            return cat, {'description': description, 'color': f'#{color.lower()}'}

        cog_categories = dict(extract_cat(cat) for cat in content.split('\n'))

        with open(file, 'w') as f:
            json.dump(cog_categories, f, indent=4)

        return cog_categories


def _get_cache_json(cache_file: str) -> dict:
    try:
        with open(cache_file) as f:
            return json.load(f)
    except Exception:
        return {}


def _set_cache_json(cache_file: str, content: Union[dict, list]) -> None:
    try:
        with open(cache_file, 'w') as f:
            json.dump(content, f)
    except Exception:
        logging.warning(f'Failed to write {content=} to {cache_file=}')


def entrez_organism_to_taxid(organism: str) -> int:
    cache_file = os.path.join(TMPDIR, 'organism_to_taxid.json')

    taxid_cache = _get_cache_json(cache_file)

    if organism in taxid_cache:
        logging.info(f'loaded taxid of {organism} from cache')
        return taxid_cache[organism]
    else:
        logging.info(f'loading taxid of {organism} from Entrez')
        handle = Entrez.esearch(db='Taxonomy', term=organism)
        record = Entrez.read(handle)
        assert 'ErrorList' not in record, f'Failed to extract taxid of {organism=} using Entrez!'
        taxids = record["IdList"]
        assert len(taxids) == 1 and taxids[0].isdigit(), f'Failed to extract taxid of {organism=} using Entrez!'
        taxid = int(taxids[0])

        taxid_cache[organism] = taxid
        _set_cache_json(cache_file, taxid_cache)
        return taxid


def clean_locus_tag(locus_tag: str) -> (str):
    return locus_tag.rsplit('|', 1)[-1]


def split_locus_tag(locus_tag: str) -> (str, str):
    locus_tag = clean_locus_tag(locus_tag)
    prefix = locus_tag.rstrip(digits)
    assert len(prefix) < len(locus_tag), f'Failed to detect {prefix=} from {locus_tag=}. Locus tags must end in digits'
    return prefix, locus_tag[len(prefix):]


def create_replace_function(replace_map: {str: str}) -> Callable:
    '''
    Returns a function that will replace all replace_map.keys with their corresponding replace_map.values

    :param replace_map: dictionary that maps strings to be replaced to their desired replacement
    :rtype: Callable
    :return: function that takes str and returns str
    '''
    replace_map = {re.escape(k): v for k, v in replace_map.items()}  # escape key
    pattern = re.compile("|".join(replace_map.keys()))

    def replace_fn(text: str) -> str:
        return pattern.sub(lambda m: replace_map[re.escape(m.group(0))], text)

    return replace_fn


def decompress_gz(gz: str, out: str):
    with gzip.open(gz, 'rb') as f_in, open(out, 'w') as f_out:
        for line in f_in:
            f_out.write(line.decode('utf-8'))


def merge_json(dict_: dict, new: str = None) -> dict:
    if new is not None and os.path.isfile(new):
        with open(new) as f:
            dict_.update(json.load(f))
    return dict_


def get_folder_structure_version(folder_structure_dir: str) -> int:
    """
    Determine current folder structure version. (Read folder_structure/version.json)

    :param folder_structure_dir: Path to the root of the OpenGenomeBrowser folder structure. (Must contain 'organisms' folder.)
    :return: version (integer)
    """
    assert type(folder_structure_dir) is str
    version_file = f'{folder_structure_dir}/version.json'

    if not os.path.isfile(version_file):
        with open(version_file, 'w') as f:
            json.dump({'folder_structure_version': 1}, f, indent=4)

    try:
        with open(version_file) as f:
            version_dict = json.load(f)
    except Exception as e:
        raise AssertionError(f'Failed to {version_file} as json: {str(e)}')

    assert 'folder_structure_version' in version_dict, f"Key 'folder_structure_version' missing in {version_file}"
    version = version_dict['folder_structure_version']
    assert type(version) is int, f"Key 'folder_structure_version' in {version_file} must be an integer!"

    return version



from Bio import SeqIO, SeqRecord, SeqFeature
from .utils import GenomeFile, query_int, entrez_organism_to_taxid, date_to_string, datetime, create_replace_function, \
    split_locus_tag
from .genbank_to_fasta import GenBankToFasta


class GenBankFile(GenomeFile):
    def rename(
            self,
            out: str,
            new_locus_tag_prefix: str,
            old_locus_tag_prefix: str = None,
            validate: bool = False,
            scf_prefix: str = None,
            scf_leading_zeroes: int = None,
    ) -> None:
        old_locus_tag_prefix = self._pre_rename_check(out, new_locus_tag_prefix, old_locus_tag_prefix)

        with open(self.path) as in_f:
            content = in_f.readlines()

        if scf_prefix:
            # change scf prefixes
            if type(scf_leading_zeroes) is int and scf_leading_zeroes > 1:
                format = lambda c: f'VERSION     {scf_prefix}{str(c).zfill(scf_leading_zeroes)}\n'
            else:
                format = lambda c: f'VERSION     {scf_prefix}{c}\n'

            counter = 0
            for i, line in enumerate(content):
                if line.startswith('VERSION'):
                    counter += 1
                    content[i] = format(counter)

        content = ''.join(content)
        # change locus tags
        replace_fn = create_replace_function({
            string.format(prefix=old_locus_tag_prefix): string.format(prefix=new_locus_tag_prefix)
            for string in ['/locus_tag="{prefix}', '/protein_id="extdb:{prefix}', ':{prefix}']
        })

        old_hash = hash(content)
        content = replace_fn(content)
        assert old_hash != hash(content), f'The content of {self.path=} has not changed!'

        assert new_locus_tag_prefix in content, f'Something went wrong: did not replace anything!'

        with open(out, 'w') as out_f:
            out_f.write(content)

        self.path = out

        if validate:
            self.validate_locus_tags(locus_tag_prefix=new_locus_tag_prefix)

    def create_ffn(self, ffn: str):
        GenBankToFasta.convert(gbk=self.path, out=ffn, format='ffn')

    def create_faa(self, faa: str):
        GenBankToFasta.convert(gbk=self.path, out=faa, format='faa')

    def validate_locus_tags(self, locus_tag_prefix: str = None):
        if locus_tag_prefix is None:
            locus_tag_prefix = self.detect_locus_tag_prefix()

        with open(self.path) as f:
            for rec in SeqIO.parse(f, "genbank"):
                for feature in rec.features:
                    locus_tag = feature.qualifiers.get('locus_tag')
                    if locus_tag is not None:
                        locus_tag = locus_tag[0]
                        real_locus_tag_prefix, gene_id = split_locus_tag(locus_tag)
                        assert real_locus_tag_prefix == locus_tag_prefix, \
                            f'locus_tag_prefix in {self.path=} does not match. expected: {locus_tag_prefix} reality: {real_locus_tag_prefix}'
                        assert gene_id.isdigit(), f'locus_tag in {self.path=} is malformed. expected: {locus_tag_prefix}_[0-9]+ reality: {locus_tag}'

    def metadata(self) -> (dict, dict):
        organism_data, genome_data = {}, {}

        try:
            organism_data['taxid'] = self.taxid(raise_error=True)
        except Exception:
            pass

        rec, feature = self._get_first_gbk_rec_feature(gbk=self.path)
        assert 'comment' in rec.annotations, f'Could not determine annotation information from {self.path=}'
        comment = rec.annotations['comment']
        if 'prokka' in comment:
            genome_data.update(
                cds_tool='prokka',
                cds_tool_date=date_to_string(datetime.strptime(rec.annotations['date'], '%d-%b-%Y')),
                cds_tool_version=comment.split('prokka ')[-1].split(' ')[0]
            )
        elif 'PGAP' in comment:
            pgap_comment = rec.annotations['structured_comment']['Genome-Annotation-Data']
            genome_data.update(
                cds_tool='PGAP',
                cds_tool_date=date_to_string(datetime.strptime(pgap_comment['Annotation Date'], '%m/%d/%Y %H:%M:%S')),
                cds_tool_version=pgap_comment['Annotation Software revision']
            )
        else:
            raise AssertionError(f'Failed to discover annotation information from {self.path=}')

        # get  BioProject / BioSample metadata
        if hasattr(rec, 'dbxrefs'):
            for dbxref in rec.dbxrefs:
                if dbxref.startswith('BioProject:'):
                    genome_data['bioproject_accession'] = dbxref.removeprefix('BioProject:')
                if dbxref.startswith('BioSample:'):
                    genome_data['biosample_accession'] = dbxref.removeprefix('BioSample:')

        return organism_data, genome_data

    def taxid(self, raise_error=True, sample_name: str = None) -> int:
        rec, feature = self._get_first_gbk_rec_feature(gbk=self.path)

        # PGAP files contain taxid in first feature (db_xref qualifier)
        db_xref = feature.qualifiers.get('db_xref')
        if type(db_xref) is list and len(db_xref) == 1:
            taxid = db_xref[0].removeprefix('taxon:')
            assert taxid.isdigit(), f'Failed to extract taxid from db_xref in {self.path=}'
            return int(taxid)

        # prokka files contain organism name, which can be turned into taxid using Entrez
        organism = feature.qualifiers.get('organism')
        if type(organism) is list and len(organism) == 1:
            return entrez_organism_to_taxid(organism[0])

        # ask for input or raise error
        if not raise_error:
            taxid = query_int(
                question=f'What is the taxid of {sample_name if sample_name else self.path}?',
                error_msg=f'The response must be an integer. Hint: taxid=1427524 is "mixed sample".'
            )
            assert taxid != 0, f'0 is not a valid taxid.'
            return taxid
        else:
            raise AssertionError(f'Failed to extract taxid from {self.path=}')

    def detect_locus_tag_prefix(self) -> str:
        strain, locus_tag_prefix = self.detect_strain_locus_tag_prefix()
        return locus_tag_prefix

    def detect_strain_locus_tag_prefix(self) -> (str, str):
        strain, locus_tag = None, None
        with open(self.path) as f:
            for rec in SeqIO.parse(f, "genbank"):
                for feature in rec.features:
                    if strain is None:
                        strain = feature.qualifiers.get('strain')
                    if locus_tag is None:
                        locus_tag = feature.qualifiers.get('locus_tag')
                    if strain is not None and locus_tag is not None:
                        break

        assert type(locus_tag) is list, f'Could not read genome from .gbk file! {locus_tag=}'
        assert type(strain) is list, f'Could not read organism from .gbk file! {strain=}'

        locus_tag_prefix, gene_id = split_locus_tag(locus_tag[0])
        strain = strain[0]
        assert type(locus_tag_prefix) is str and type(strain) is str
        return strain, locus_tag_prefix

    @staticmethod
    def _get_first_gbk_rec_feature(gbk: str) -> (SeqRecord, SeqFeature):
        with open(gbk) as f:
            for rec in SeqIO.parse(f, "genbank"):
                for feature in rec.features:
                    return rec, feature
        raise AssertionError(f'Failed to get rec and feature from {gbk=}')


def rename_genbank(
        file: str, out: str,
        new_locus_tag_prefix: str, old_locus_tag_prefix: str = None, validate: bool = False,
        scf_prefix: str = None, scf_leading_zeroes: int = None,
):
    """
    Change the locus tags in a GenBank file

    :param file: input file
    :param out: output file
    :param new_locus_tag_prefix: desired locus tag
    :param old_locus_tag_prefix: locus tag to replace
    :param validate: if true, perform sanity check on locus tags
    :param scf_prefix: desired scaffold prefix (optional)
    :param scf_leading_zeroes: format scaffold counter with leading zeroes. e.g.: 5 -> >PREFIX_00001 (optional)
    """

    GenBankFile(
        file=file
    ).rename(
        out=out,
        new_locus_tag_prefix=new_locus_tag_prefix,
        old_locus_tag_prefix=old_locus_tag_prefix,
        validate=validate,
        scf_prefix=scf_prefix,
        scf_leading_zeroes=scf_leading_zeroes
    )


def main():
    import fire

    fire.Fire(rename_genbank)


if __name__ == '__main__':
    main()



import os
import json
import shutil
from urllib import request
from .utils import PACKAGE_ROOT
from . import __folder_structure_version__


def download_go_data(out: str) -> None:
    source_url = 'http://purl.obolibrary.org/obo/go.obo'

    print(f'Converting {source_url} -> {out}')

    def go_generator(io) -> [str]:
        go_entry = []

        line = io.readline()
        while line:
            if line == b'[Term]\n':
                yield go_entry
                go_entry.clear()

            go_entry.append(line.decode('utf-8'))
            line = io.readline()

        yield go_entry

    def get_name(entry: list) -> str:
        for line in entry:
            if line.startswith('name: '):
                return line.rstrip()[6:]
        raise TypeError(F'The go.obo file seems to have a wrong format! broken entry: {entry}')

    def get_go(entry: list) -> str:
        entry = entry[1]
        assert entry.startswith('id: GO:') and len(entry) == 15, f'Bad entry in go.obo: {entry}, len={len(entry)}'
        assert entry[7:14].isnumeric()
        return entry[4:14]

    with request.urlopen(source_url) as source_handle, open(out, 'w') as target_handle:
        gos = go_generator(io=source_handle)

        # skip first entry
        file_head = next(gos)
        assert not file_head[0].startswith('[Term]'), F'The go.obo file seems to have a wrong format! file_head looks wrong: {file_head}'

        # save regular entries to file
        for entry in gos:
            target_handle.write(F'{get_go(entry)}\t{get_name(entry)}\n')


def download_kegg_data(src: str, out: str, remove_prefix: str = '', add_prefix: str = '') -> None:
    source_url = f'http://rest.kegg.jp/list/{src}'

    print(f'Converting {source_url} -> {out}')

    with request.urlopen(source_url) as source_handle, open(out, 'w') as target_handle:
        for line in source_handle:
            target_handle.write(f'{add_prefix}{line.decode("utf-8").removeprefix(remove_prefix)}')


def download_sl_data(out: str) -> None:
    source_url = 'https://www.uniprot.org/locations/?query=*&format=tab&force=true&columns=id'

    print(f'Converting {source_url} -> {out}')

    error_msg = 'UniProt must have changed its format. Please contact the developer. error={error}'

    with request.urlopen(source_url) as source_handle, open(out, 'w') as target_handle:
        first_line = source_handle.readline().decode('utf-8')
        assert first_line == 'Subcellular location ID\tDescription\tCategory\tAlias\n', error_msg.format(error=first_line)

        for line in source_handle:
            line = line.decode('utf-8').strip().split('\t')
            assert len(line) == 4, error_msg.format(error=f'{len(line)=}; {line=}')
            sl, description, type, location = line

            target_handle.write(f'{sl}\t{location} ({description})\n')


def init_folder_structure(folder_structure_dir: str = None) -> None:
    """
    Creates a basic OpenGenomeBrowser folders structure.

    Result:
        folder_structure
        ├── organisms
        ├── annotations.json
        ├── annotation-descriptions
        │   ├── SL.tsv
        │   ├── KO.tsv
        │   ├── KR.tsv
        │   ├── EC.tsv
        │   └── GO.tsv
        ├── orthologs
        └── pathway-maps
            ├── type_dictionary.json
            └── svg


    :param folder_structure_dir: Path to the root of the OpenGenomeBrowser folder structure. (Will contain 'organisms' folder.)
    """
    if folder_structure_dir is None:
        assert 'FOLDER_STRUCTURE' in os.environ, \
            f'Cannot find the folder_structure. ' \
            f'Please set --folder_structure_dir or environment variable FOLDER_STRUCTURE'
        folder_structure_dir = os.environ['FOLDER_STRUCTURE']

    assert os.path.isdir(os.path.dirname(folder_structure_dir)), f'Parent dir of {folder_structure_dir=} does not exist!'
    assert not os.path.exists(folder_structure_dir), f'Error: {folder_structure_dir=} already exist!'

    # make main dir
    os.makedirs(folder_structure_dir)

    # set version
    with open(f'{folder_structure_dir}/version.json', 'w') as f:
        json.dump({'folder_structure_version': __folder_structure_version__}, f, indent=4)

    # make organisms dir (empty)
    os.makedirs(f'{folder_structure_dir}/organisms')

    # make orthologs dir (empty)
    os.makedirs(f'{folder_structure_dir}/orthologs')

    # make pathway maps dir and content
    os.makedirs(f'{folder_structure_dir}/pathway-maps')
    os.makedirs(f'{folder_structure_dir}/pathway-maps/svg')
    with open(f'{folder_structure_dir}/pathway-maps/type_dictionary.json', 'w') as f:
        f.write('{}')

    # Create annotations.json
    shutil.copy(src=f'{PACKAGE_ROOT}/data/annotations.json', dst=f'{folder_structure_dir}/annotations.json')

    # download annotation descriptions
    annotation_descriptions_dir = f'{folder_structure_dir}/annotation-descriptions'
    os.makedirs(annotation_descriptions_dir)
    download_sl_data(out=f'{annotation_descriptions_dir}/SL.tsv')
    download_kegg_data(src='rn', out=f'{annotation_descriptions_dir}/KR.tsv', remove_prefix='rn:')
    download_kegg_data(src='ko', out=f'{annotation_descriptions_dir}/KG.tsv', remove_prefix='ko:')
    download_kegg_data(src='enzyme', out=f'{annotation_descriptions_dir}/EC.tsv', remove_prefix='ec:', add_prefix='EC:')
    download_go_data(out=f'{annotation_descriptions_dir}/GO.tsv')


def main():
    import fire

    fire.Fire(init_folder_structure)


if __name__ == '__main__':
    main()



import os
import json
import logging

from .folder_looper import FolderLooper, FolderGenome
from .rename_eggnog import EggnogFile
from .utils import query_yes_no, get_folder_structure_version


def _get_folder_structure_dir(folder_structure_dir: str = None) -> str:
    if folder_structure_dir is None:
        assert 'FOLDER_STRUCTURE' in os.environ, f'Cannot find the folder_structure. Please set --folder_structure_dir or environment variable FOLDER_STRUCTURE'
        folder_structure_dir = os.environ['FOLDER_STRUCTURE']
    assert os.path.isdir(folder_structure_dir), f'Could not find the folder_structure. Folder does not exist: {folder_structure_dir}'
    return folder_structure_dir


def set_folder_structure_version(new_version: int, folder_structure_dir: str) -> None:
    assert type(folder_structure_dir) is str
    version_file = f'{folder_structure_dir}/version.json'

    with open(version_file) as f:
        version_dict = json.load(f)

    version_dict['folder_structure_version'] = new_version

    with open(version_file, 'w') as f:
        json.dump(version_dict, f, indent=4)

    print()
    print(f'Successfully updated to folder structure version {new_version}!')


def ask(v_from: int, v_to: int, actions: [str], folder_structure_dir: str):
    assert type(folder_structure_dir) is str

    current_version = get_folder_structure_version(folder_structure_dir)
    assert current_version == v_from, \
        f'Cannot proceed: Folder structure version mismatch.\n' \
        f'This script expects version {v_from}, but folder_structure/version.json says version {current_version}.'

    question = f'Upgrade folder structure from version {v_from} to {v_to}:'
    for action in actions:
        question += f'\n - {action}'
    question += '\n\nProceed?'
    if not query_yes_no(question=question, default='yes'):
        exit(1)


def loop_genomes(folder_structure_dir: str, skip_ignored=False, sanity_check=False, representatives_only=False) -> [FolderGenome]:
    for genome in FolderLooper(folder_structure_dir=folder_structure_dir).genomes(
            skip_ignored=skip_ignored,
            sanity_check=sanity_check,
            representatives_only=representatives_only
    ):
        if genome.has_json:
            yield genome


def from_1_to_2(folder_structure_dir: str = None, skip_ignored=False, sanity_check=False, representatives_only=False):
    """ Upgrade OpenGenomeBrowser folder structure. """
    folder_structure_dir = _get_folder_structure_dir(folder_structure_dir)
    v_from = 1
    v_to = 2

    ask(v_from=v_from, v_to=v_to, actions=['add COG to genome.json'], folder_structure_dir=folder_structure_dir)

    for genome in loop_genomes(folder_structure_dir=folder_structure_dir, skip_ignored=skip_ignored, sanity_check=sanity_check,
                               representatives_only=representatives_only):
        genome_json = genome.json
        if 'COG' in genome_json:
            print(f'{genome.identifier}: already has COG in genome.json')
            continue

        COG = {}  # default

        eggnog_files = [f for f in genome_json['custom_annotations'] if f['type'].startswith('eggnog')]
        for file in eggnog_files:
            path = os.path.join(genome.path, file['file'])
            try:
                COG = EggnogFile(file=path).cog_categories()
            except AssertionError as e:
                logging.info(msg=str(e))
                pass

        print(f'{genome.identifier}: adding COG={COG}')
        genome_json['COG'] = COG
        genome.replace_json(genome_json)

    set_folder_structure_version(new_version=v_to, folder_structure_dir=folder_structure_dir)


def main():
    from fire import Fire

    Fire({
        'get_current_version': get_folder_structure_version,
        '1_to_2': from_1_to_2,
    })


if __name__ == '__main__':
    main()



import json
from functools import cached_property
from json.decoder import JSONDecodeError
import os
import shutil
from datetime import datetime
from .metadata_schemas import organism_json_schema, genome_json_schema


def set_to_list(obj):
    if isinstance(obj, set):
        return list(obj)
    raise TypeError(f'Could not serialize {obj}')


class FolderEntity:
    path: str
    json_path: str

    def __init__(self, path: str):
        assert os.path.isdir(path), path
        self.path = path

    @property
    def is_ignored(self):
        return os.path.isfile(f'{self.path}/ignore')

    @property
    def has_json(self):
        return self.json is not None

    @cached_property
    def json(self):
        try:
            with open(self.json_path) as f:
                return json.load(f)
        except (FileNotFoundError, JSONDecodeError):
            return None

    def get_json_attr(self, attr: str):
        assert self.has_json, f'Could not extract {attr=} from {self.json_path}.'
        return self.json[attr]

    def sanity_check(self):
        if self.is_ignored:
            return True
        else:
            assert self.has_json, f'{self} :: does not have a (valid) json'

    def replace_json(self, data):
        assert type(data) is dict
        # ensure data is serializable
        try:
            json.dumps(data, default=set_to_list)
        except json.JSONDecodeError as e:
            raise AssertionError(f'Could not save dictionary as json: {e}')

        date = datetime.now().strftime("%Y_%b_%d_%H_%M_%S")

        # create backup
        bkp_dir = F'{self.path}/.bkp'
        os.makedirs(bkp_dir, exist_ok=True)
        bkp_file = F'{bkp_dir}/{date}_folderlooper_organism.json'
        shutil.move(src=self.json_path, dst=bkp_file)

        # write new file
        assert not os.path.isfile(self.json_path)
        with open(self.json_path, 'w') as f:
            json.dump(data, f, sort_keys=True, indent=4, default=set_to_list)


class FolderOrganism(FolderEntity):
    def __init__(self, path: str):
        super().__init__(path)
        self.name = os.path.basename(self.path)
        self.json_path = self.path + "/organism.json"

    def __str__(self):
        return f'<Organism {self.name}>'

    def sanity_check(self):
        super().sanity_check()

        assert self.name == self.get_json_attr('name'), \
            f"{self} :: 'name' in organism.json doesn't match folder name: {self.path}"

        organism_json_schema.validate(self.json)

        representative = self.representative()

        assert representative.identifier.startswith(self.name), \
            f'{self} :: representative identifer must start with organism name'
        assert os.path.isdir(representative.path), \
            F"{self} :: Representative doesn't exist! Representative: {representative}"
        assert not representative.is_ignored, \
            F"{self} :: Representatives may not be ignored! Representative: {representative}, {representative.path}, {representative.is_ignored}"

    @property
    def genomes_path(self):
        return f'{self.path}/genomes'

    @property
    def representative_path(self):
        return f'{self.path}/genomes/{self.get_json_attr("representative")}'

    def representative(self, sanity_check=True):
        """returns: Genome"""
        representative = FolderGenome(path=self.representative_path, organism=self)
        if sanity_check:
            representative.sanity_check()
        return representative

    def genomes(self, skip_ignored: bool, sanity_check=True) -> []:
        """generator, yields [Genome]"""
        genomes_folders = os.scandir(self.genomes_path)
        for genome_folder in genomes_folders:
            genome = FolderGenome(path=genome_folder.path, organism=self)
            if skip_ignored and genome.is_ignored:
                continue
            if sanity_check:
                genome.sanity_check()
            yield genome


class FolderGenome(FolderEntity):
    def __init__(self, path: str, organism: FolderOrganism):
        super().__init__(path)
        self.identifier = os.path.basename(self.path)
        self.organism = organism
        self.json_path = self.path + "/genome.json"

    def __str__(self):
        return f'<Genome {self.identifier}>'

    def sanity_check(self):
        super().sanity_check()

        assert self.identifier == self.get_json_attr('identifier'), \
            F"{self} :: 'identifier' in genome.json doesn't match folder name: {self.path}"

        assert self.identifier.startswith(self.organism.name), f'{self} :: identifer must start with organism name'

        genome_json_schema.validate(self.json)


class FolderLooper:
    def __init__(self, folder_structure_dir: str):
        self.folder_structure_dir = folder_structure_dir
        self.organism_path = os.path.join(self.folder_structure_dir, 'organisms')

        for dir in (self.folder_structure_dir, self.organism_path):
            assert os.path.isdir(dir), f'Folder does not exist: {dir=}'

    def organisms(self, skip_ignored: bool = True, sanity_check: bool = True) -> [FolderOrganism]:
        """generator"""
        for organism_folder in os.scandir(self.organism_path):
            organism = FolderOrganism(path=organism_folder.path)
            if skip_ignored and organism.is_ignored:
                continue
            if sanity_check:
                organism.sanity_check()
            yield organism

    def genomes(self, skip_ignored: bool = True, sanity_check: bool = True, representatives_only: bool = False) -> [FolderGenome]:
        """generator"""
        for organism in self.organisms(skip_ignored=skip_ignored, sanity_check=sanity_check):
            if representatives_only:
                yield organism.representative(sanity_check=sanity_check)
            else:
                for genome in organism.genomes(skip_ignored=skip_ignored, sanity_check=sanity_check):
                    yield genome


def loop(folder_structure_dir: str = None, what: str = 'genomes', skip_ignored: bool = True, sanity_check: bool = True, representatives_only: bool = False):
    if folder_structure_dir is None:
        folder_structure_dir = os.environ.get('FOLDER_STRUCTURE')

    folder_looper = FolderLooper(folder_structure_dir)
    if what == 'genomes':
        for organism in folder_looper.organisms(skip_ignored=skip_ignored, sanity_check=sanity_check):
            print(organism.path)
    elif what == 'organisms':
        for genome in folder_looper.genomes(skip_ignored=skip_ignored, sanity_check=sanity_check, representatives_only=representatives_only):
            print(genome.path)
    else:
        raise AssertionError(f"{what=} must be either 'genomes' or 'organisms'.")


def main():
    import fire

    fire.Fire(loop)


if __name__ == '__main__':
    main()



from re import compile
from datetime import datetime

extr_digits = compile('[0-9]+')


def parse_busco(file: str) -> dict:
    entries = [
        ('Complete BUSCOs (C)', 'C'),
        ('Complete and single-copy BUSCOs (S)', 'S'),
        ('Complete and duplicated BUSCOs (D)', 'D'),
        ('Fragmented BUSCOs (F)', 'F'),
        ('Missing BUSCOs (M)', 'M'),
        ('Total BUSCO groups searched', 'T')
    ]

    busco_dict = {"C": None, "D": None, "F": None, "M": None, "S": None, "T": None}

    error_msg = F'Error while parsing BUSCO: {file} - {busco_dict}'

    with open(file) as f:
        busco_file = f.readlines()

    for line in busco_file:
        for descr, abbr in entries:
            if descr in line:
                x = extr_digits.findall(line)
                assert len(x) == 1, error_msg
                busco_dict[abbr] = int(x[0])

    assert len(busco_dict.keys()) == 6 and None not in busco_dict.values(), error_msg
    assert busco_dict['C'] + busco_dict['M'] + busco_dict['F'] == busco_dict['T'], error_msg
    assert busco_dict['S'] + busco_dict['D'] == busco_dict['C'], error_msg

    dataset_line = [l for l in busco_file if l.startswith('# The lineage dataset is:')][0]
    dataset, rest = dataset_line[26:].split(' (Creation date: ', maxsplit=1)
    dataset_date = rest[:10]
    try:
        datetime.strptime(dataset_date, '%Y-%m-%d')
    except Exception as e:
        raise AssertionError(F'Bad busco creation date ! {file}n{str(e)}')

    busco_dict['dataset'] = dataset
    busco_dict['dataset_creation_date'] = dataset_date

    return busco_dict


def main():
    import fire
    import json

    def runner(file: str):
        print(json.dumps(parse_busco(file), indent=4))

    fire.Fire(runner)


if __name__ == '__main__':
    main()



import os
import pandas as pd
from Bio import SeqIO
from collections import Counter
from .utils import clean_locus_tag


class OrthogroupToGeneName:
    def __init__(self, fasta_dir: str, file_endings='fasta'):
        assert os.path.isdir(fasta_dir), F'fasta_dir does not exist: "{fasta_dir}"'
        self.fasta_dir = fasta_dir
        self.file_endings = file_endings

    def majority_dict(self):
        assert hasattr(self, 'majority_df'), F'Load Orthogroups.tsv or N0.tsv first!'
        return {
            orthogroup: majority_name
            for orthogroup, majority_name
            in self.majority_df['Best Gene Name'].iteritems()
        }

    def save_majority_df(self, out_file: str):
        """
        Writes the following file:

        HOG Best Gene Name  Gene Name Occurrences
        N0.HOG0000000   amino acid ABC transporter Counter({'amino acid ABC transporter': 43})
        ...
        """
        assert hasattr(self, 'majority_df'), F'Load Orthogroups.tsv or N0.tsv first!'
        self.majority_df.to_csv(path_or_buf=out_file, sep='\t')

    def save_orthogroup_to_gene_ids(self, out_file: str):
        """
        Writes the following file (no header):

        N0.HOG0000000   gene_1, gene_2
        N0.HOG0000001   gene_3, gene_4, gene_5
        ...
        """
        with open(out_file, 'w') as f:
            for orthogroup, row in self.gene_ids_df.iterrows():
                gene_ids = [gene_id for gene_ids in row for gene_id in gene_ids]
                gene_ids = ', '.join(gene_ids)
                f.write(F'{orthogroup}\t{gene_ids}\n')

    def save_orthogroup_to_best_name(self, out_file: str):
        """
        Writes the following file (no header):

        N0.HOG0000000	amino acid ABC transporter ATP-binding protein
        N0.HOG0000001	ATP-binding cassette domain-containing protein
        ...
        """

        self.majority_df.drop(
            columns=['Gene Name Occurrences'],
            inplace=False
        ).to_csv(
            path_or_buf=out_file,
            sep='\t',
            header=False
        )

    def load_og(self, og_tsv: str):
        assert os.path.isfile(og_tsv), F'og_tsv does not exist: "{og_tsv}"'

        # read "Orthogroups.tsv"
        gene_ids_df = pd.read_csv(og_tsv, sep='\t', dtype=str)
        gene_ids_df.set_index('Orthogroup', inplace=True)
        gene_ids_df = gene_ids_df.applymap(
            lambda x: [] if pd.isnull(x) else [clean_locus_tag(locus_tag) for locus_tag in x.split(', ')])

        self.gene_ids_df = gene_ids_df
        self.__load_gene_names(self.gene_ids_df.__deepcopy__())

    def load_hog(self, n0_tsv: str):
        assert os.path.isfile(n0_tsv), F'n0_tsv does not exist: "{n0_tsv}"'

        # read "N0.tsv" and drop extra columns
        gene_ids_df = pd.read_csv(n0_tsv, sep='\t', dtype=str)
        gene_ids_df.set_index('HOG', inplace=True)
        gene_ids_df.drop(columns=['OG', 'Gene Tree Parent Clade'], inplace=True)
        gene_ids_df = gene_ids_df.applymap(
            lambda x: [] if pd.isnull(x) else [clean_locus_tag(locus_tag) for locus_tag in x.split(', ')])

        self.gene_ids_df = gene_ids_df
        self.__load_gene_names(self.gene_ids_df.__deepcopy__())

    def __load_gene_names(self, gene_names_df: pd.DataFrame):
        self.strains = gene_names_df.columns

        for strain in self.strains:
            gene_id_to_name = self.__get_gene_id_to_name_dict(strain)
            gene_names_df[strain] = gene_names_df[strain].apply(lambda ids: [gene_id_to_name[id] for id in ids])

        self.majority_df = pd.DataFrame(index=gene_names_df.index)
        self.majority_df[['Best Gene Name', 'Gene Name Occurrences']] = gene_names_df.apply(
            lambda row: pd.Series(self.__majority_vote(row)), axis=1
        )

    def __majority_vote(self, row):
        """ Get gene names set per HOG and count them. Disregards names with eAED in description,
        typically useless maker annotations"""

        all_names = [name for cell in row for name in cell if "eAED" not in name]
        if len(all_names) == 0:
            all_names = ["NO DESCRIPTION"]  # if all names contain "eAED", all_names is an empty list

        names_to_count = Counter(all_names)

        return names_to_count.most_common(1)[0][0], str(
            names_to_count)  # return best gene name and gene name occurrences as string

    def __get_gene_id_to_name_dict(self, strain):
        fasta_file_path = os.path.join(self.fasta_dir + f'/{strain}.{self.file_endings}')
        assert os.path.isfile(fasta_file_path), F'{self.file_endings} file "{fasta_file_path}" is missing!'
        genes = SeqIO.parse(fasta_file_path, "fasta")

        def extract_description(gene):
            description = gene.description.split(' ', maxsplit=1)
            assert len(description) == 2, \
                F'Failed to extract description for strain={strain}:\n' \
                F'gene.id={gene.id}\n' \
                F'gene.description={gene.description}\n' \
                F'description={description}'
            description = description[1]
            if description.endswith(']'):
                # remove species description
                return description.rsplit(' [', maxsplit=1)[0]
            else:
                return description

        return {clean_locus_tag(gene.id): extract_description(gene) for gene in genes}


def import_orthofinder(
        folder_structure_dir: str = None,
        fasta_dir: str = None,
        out_annotations: str = None,
        out_descriptions: str = None,
        which: str = 'hog'
):
    """
    Creates two files:
      - folder_structure/orthologs/orthologs.tsv
      - folder_structure/annotation-descriptions/OL.tsv

    :param folder_structure_dir: Path to the root of the OpenGenomeBrowser folder structure. (Must contain 'organisms' folder.)
    :param fasta_dir: Path to the OrthoFinder FASTA dir
    :param out_annotations: Path to output annotation file (maps OG -> genes)
    :param out_descriptions: Path to output descriptions file (maps OG -> description)
    :param which: either 'hog' for N0.tsv or 'og' for 'Orthogroups.tsv
    """
    if folder_structure_dir is None:
        assert 'FOLDER_STRUCTURE' in os.environ, f'Cannot find the folder_structure. Please set --folder_structure_dir or environment variable FOLDER_STRUCTURE'
        folder_structure_dir = os.environ['FOLDER_STRUCTURE']

    if fasta_dir is None:
        fasta_dir = f'{folder_structure_dir}/OrthoFinder/fastas'

    if out_descriptions is None:
        out_descriptions = f'{folder_structure_dir}/annotation-descriptions/OL.tsv'  # is AnnotationDescriptionFile, maps OG -> description

    if out_annotations is None:
        out_annotations = f'{folder_structure_dir}/orthologs/orthologs.tsv'  # is CustomAnnotationsFile, maps HOG -> [gene]

    assert os.path.isdir(fasta_dir), f'Error: directory does not exist: {fasta_dir=}'
    assert os.path.isdir(
        f'{fasta_dir}/OrthoFinder'), f'Error: {fasta_dir=} contains no OrthoFinder directory. You must run OrthoFinder first!'
    for file in (out_descriptions, out_annotations):
        assert not os.path.isfile(file), f'Output file already exists! {file=}'

    results_dir = os.listdir(f'{fasta_dir}/OrthoFinder')
    assert len(
        results_dir) == 1, f'Error: {fasta_dir}/OrthoFinder must contain exactly one directory! dirs={results_dir}'
    results_dir = results_dir[0]

    og_tsv = f'{fasta_dir}/OrthoFinder/{results_dir}/Orthogroups/Orthogroups.tsv'
    hog_tsv = f'{fasta_dir}/OrthoFinder/{results_dir}/Phylogenetic_Hierarchical_Orthogroups/N0.tsv'

    otg = OrthogroupToGeneName(fasta_dir, file_endings='faa')

    if which == 'hog':
        assert os.path.isfile(hog_tsv), f'Could not find {hog_tsv=}'
        otg.load_hog(n0_tsv=hog_tsv)
    elif which == 'og':
        assert os.path.isfile(og_tsv), f'Could not find {og_tsv=}'
        otg.load_og(og_tsv=og_tsv)
    else:
        raise AssertionError(f"which must be 'og' or 'hog'")

    otg.save_orthogroup_to_gene_ids(out_annotations)
    otg.save_orthogroup_to_best_name(out_descriptions)

    print(f'Sucessfully generated {out_annotations} and {out_descriptions}')


def main():
    import fire

    fire.Fire(import_orthofinder)


if __name__ == '__main__':
    main()



from re import compile
from datetime import datetime
from functools import cached_property

from .utils import GenomeFile, split_locus_tag, get_cog_categories

EGGNOG_VERSIONS = {
    'eggnog-2.1.2':
        '#query\tseed_ortholog\tevalue\tscore\teggNOG_OGs\tmax_annot_lvl\tCOG_category\tDescription\tPreferred_name\tGOs\tEC\tKEGG_ko\tKEGG_Pathway\tKEGG_Module\tKEGG_Reaction\tKEGG_rclass\tBRITE\tKEGG_TC\tCAZy\tBiGG_Reaction\tPFAMs\n',
    'eggnog':
        '#query_name\tseed_eggNOG_ortholog\tseed_ortholog_evalue\tseed_ortholog_score\tbest_tax_level\tPreferred_name\tGOs\tEC\tKEGG_ko\tKEGG_Pathway\tKEGG_Module\tKEGG_Reaction\tKEGG_rclass\tBRITE\tKEGG_TC\tCAZy\tBiGG_Reaction\n',
}


class EggnogFile(GenomeFile):
    def rename(self, out: str, new_locus_tag_prefix: str, old_locus_tag_prefix: str = None, validate: bool = False) -> None:
        old_locus_tag_prefix = self._pre_rename_check(out, new_locus_tag_prefix, old_locus_tag_prefix)

        with open(self.path) as in_f:
            content = in_f.readlines()

        def rename_line(line: str):
            if line.startswith('#'):
                return line

            locus_tag, rest = line.split('\t', 1)
            locus_tag_prefix, gene_id = split_locus_tag(locus_tag)
            assert locus_tag_prefix == old_locus_tag_prefix, f'Eggnog line does not contain old_locus_tag_prefix!' \
                                                             f'{old_locus_tag_prefix=}, {line=}, {self.path=}'
            return locus_tag.replace(old_locus_tag_prefix, new_locus_tag_prefix, 1) + '\t' + rest

        content = [rename_line(line) for line in content]

        with open(out, 'w') as out_f:
            out_f.writelines(content)

        self.path = out

        if validate:
            self.validate_locus_tags(locus_tag_prefix=new_locus_tag_prefix)

    def detect_locus_tag_prefix(self) -> str:
        with open(self.path) as f:
            for line in f:
                if line.startswith('#'):
                    continue

                locus_tag = line.split('\t', 1)[0]
                locus_tag_prefix, gene_id = split_locus_tag(locus_tag)
                return locus_tag_prefix

        raise KeyError(f'Could not extract locus_tag from {self.path=}, it does not appear to contain annotations!')

    def date(self) -> datetime:
        with open(self.path) as f:
            head = [next(f) for x in range(4)]

        if head[0].startswith('##'):
            date_line = head[0].strip()
            dt = datetime.strptime(date_line[3:], '%c')
        elif head[0].startswith('#'):
            date_line = head[2].strip()
            dt = datetime.strptime(date_line[8:], '%c')
        else:
            dt = super().date()  # use creation date of file

        return dt

    @cached_property
    def custom_annotation_type(self) -> str:
        with open(self.path) as f:
            head = '\n'.join(next(f) for x in range(5))  # read 5 lines

        for type, columns_header in EGGNOG_VERSIONS.items():
            if columns_header in head:
                return type

        raise KeyError(f'Could not discover eggnog type! {self.path=}')

    def validate_locus_tags(self, locus_tag_prefix: str = None):
        if locus_tag_prefix is None:
            locus_tag_prefix = self.detect_locus_tag_prefix()

        with open(self.path) as f:
            for line in f:
                if line.startswith('#'):
                    continue

                locus_tag = line.split('\t', 1)[0]
                real_locus_tag_prefix, gene_id = split_locus_tag(locus_tag)

                assert real_locus_tag_prefix == locus_tag_prefix, \
                    f'locus_tag_prefix in {self.path=} does not match. expected: {locus_tag_prefix} reality: {real_locus_tag_prefix}'
                assert gene_id.isdigit(), f'locus_tag in {self.path=} is malformed. expected: {locus_tag_prefix}_[0-9]+ reality: {locus_tag}'

    def cog_categories(self) -> dict:
        cog_categories = get_cog_categories()

        re_categories = compile(pattern=f"-|[{''.join(cog_categories.keys())}]+")

        cog_to_count = {cat: 0 for cat in cog_categories.keys()}
        cog_to_count['-'] = 0  # eggnog assigns '-' sometimes; surely it means the same as 'S': Function unknown
        n_genes = 0

        with open(self.path) as f:
            for line in f:
                if line.startswith('#'):
                    continue
                n_genes += 1

                cog_cats = line.split('\t')[6]

                assert re_categories.fullmatch(cog_cats) is not None, f'Failed to interpret "{cog_cats}" as COG categories. {self.path=} {line=}'

                weight = 1 / len(cog_cats)
                for cog_cat in cog_cats:
                    cog_to_count[cog_cat] += weight

        # merge '-' and 'S'
        cog_to_count['S'] += cog_to_count['-']
        del cog_to_count['-']

        # remove empty categories
        cog_to_count = {cog_cat: count for cog_cat, count in cog_to_count.items() if count != 0}

        assert n_genes == round(sum(cog_to_count.values()))

        return cog_to_count


def rename_eggnog(file: str, out: str, new_locus_tag_prefix: str, old_locus_tag_prefix: str = None, validate: bool = False):
    """
    Change the locus tags in a eggnog output file (.emapper.annotations)

    :param file: input file
    :param out: output file
    :param new_locus_tag_prefix: desired locus tag
    :param old_locus_tag_prefix: locus tag to replace
    :param validate: if true, perform sanity check
    """

    EggnogFile(
        file=file
    ).rename(
        out=out,
        new_locus_tag_prefix=new_locus_tag_prefix,
        old_locus_tag_prefix=old_locus_tag_prefix,
        validate=validate
    )


def main():
    import fire

    fire.Fire(rename_eggnog)


if __name__ == '__main__':
    main()



import io
import os
import logging
from typing import Optional

from Bio import SeqIO, SeqRecord, SeqFeature


class GenBankToFasta:
    @classmethod
    def convert(cls, gbk, out: str, format: str, strict: bool = True):
        """
        Convert GenBank (gbk) file into protein FASTA (faa) or nucleotide FASTA (ffn)

        :param gbk: path to input GenBank file
        :param out: path to output file
        :param format: faa for protein FASTA, ffn for nucleotide FASTA
        """
        assert not os.path.isfile(out), f'Output file already exists! {out=}'
        assert format in ('faa', 'ffn'), f'Format must be either faa or ffn! {format=}'

        with open(out, 'w') as f:
            for line in cls._short_fasta_generator(gbk=gbk, format=format, strict=strict):
                f.write(line)

    @classmethod
    def _short_fasta_generator(cls, gbk: str, format: str, strict: bool):
        """
        Generator that turns gbk-file into standard nucleotide FASTA-format.

        The nucleotide sequence is interrupted by newlines every 50 bp.
        """

        fasta = list(cls._long_fasta_generator(gbk=gbk, format=format, strict=strict))
        fasta = '\n'.join(fasta)
        fasta = io.StringIO(fasta)
        for seq in SeqIO.parse(fasta, "fasta"):
            yield seq.format("fasta")

    @classmethod
    def _long_fasta_generator(cls, gbk: str, format: str, strict: bool):
        """
        Generator that turns gbk-file into long-form nucleotide FASTA-format.

        The nucleotide sequence is not interrupted by newlines.
        """

        if format == 'faa':
            expected_entries = cls._get_total_proteins(gbk=gbk)
            expected_type = 'proteins'

            def parse_feature(feature: SeqFeature):
                if 'locus_tag' in feature.qualifiers and 'product' in feature.qualifiers and 'translation' in feature.qualifiers:
                    locus_tag = feature.qualifiers['locus_tag'][0]
                    gene_product = feature.qualifiers['product'][0]
                    sequence = feature.qualifiers['translation'][0]
                    return locus_tag, gene_product, sequence
                else:
                    return None, None, None

        elif format == 'ffn':
            expected_entries = cls._get_total_genes(gbk=gbk)
            expected_type = 'genes'

            def parse_feature(feature: SeqFeature):
                if 'locus_tag' in feature.qualifiers and 'product' in feature.qualifiers:
                    locus_tag = feature.qualifiers['locus_tag'][0]
                    gene_product = feature.qualifiers['product'][0]
                    sequence = str(feature.location.extract(rec).seq)
                    return locus_tag, gene_product, sequence
                else:
                    return None, None, None

        else:
            raise AssertionError(f'Format must be either faa or ffn! {format=}')

        locus_tags = set()
        with open(gbk) as f:
            for rec in SeqIO.parse(f, "genbank"):
                for feature in rec.features:
                    locus_tag, gene_product, sequence = parse_feature(feature=feature)
                    if locus_tag is None:
                        continue

                    # check if locus_tag is unique
                    if locus_tag in locus_tags:
                        logging.warning(f'GenBank is strange: {locus_tag} occurs multiple times!')
                    locus_tags.add(locus_tag)

                    yield f'>{locus_tag} {gene_product}\n{sequence}'

        if expected_entries is not None and expected_entries != len(locus_tags):
            msg = f'GenBank is strange: Claims to have {expected_entries} {expected_type}, but only has {len(locus_tags)} unique locus_tags!'
            if strict:
                raise AssertionError(msg)
            else:
                logging.warning(msg)

    @classmethod
    def _get_total_genes(cls, gbk: str) -> Optional[int]:
        rec, feature = cls._get_first_gbk_rec_feature(gbk=gbk)
        try:
            return int(rec.annotations['structured_comment']['Genome-Annotation-Data']['Genes (total)'])
        except Exception:
            return None

    @classmethod
    def _get_total_proteins(cls, gbk: str) -> Optional[int]:
        rec, feature = cls._get_first_gbk_rec_feature(gbk=gbk)
        try:
            return int(rec.annotations['structured_comment']['Genome-Annotation-Data']['CDSs (with protein)'])
        except Exception:
            return None

    @classmethod
    def _get_first_gbk_rec_feature(cls, gbk: str) -> (SeqRecord, SeqFeature):
        with open(gbk) as f:
            for rec in SeqIO.parse(f, "genbank"):
                for feature in rec.features:
                    return rec, feature
        raise AssertionError(f'Failed to get rec and feature from {gbk=}')


def main():
    import fire

    fire.Fire(GenBankToFasta.convert)


if __name__ == '__main__':
    main()



import logging
import os
import re
import json
import yaml
import shutil
import tempfile
from glob import glob
from textwrap import shorten
from typing import Union, Optional
from schema import SchemaError

from . import __folder_structure_version__
from .utils import entrez_organism_to_taxid, GenomeFile, merge_json, get_folder_structure_version
from .rename_genbank import GenBankFile
from .rename_gff import GffFile
from .rename_fasta import FastaFile
from .rename_eggnog import EggnogFile
from .parse_busco import parse_busco
from .rename_custom_annotations import CustomAnnotationFile
from .metadata_schemas import \
    organism_json_schema, genome_json_dummy, genome_json_schema, organism_json_dummy


class ImportSettings:
    settings: dict[re.Pattern:str]
    default_settings = {
        'organism_template': {},
        'genome_template': {},
        'manual_paths': {
            'eggnog': r'.*\.emapper\.annotations',
            'fna': r'.*\.fna',
            'faa': r'.*\.faa',
            'gbk': r'.*\.gbk',
            'gff': r'.*\.gff',
            'sqn': r'.*\.sqn',
            'ffn': r'.*\.ffn',
            'genome.md': r'genome\.md',
            'organism.md': r'organism\.md',
            'yaml': r'.*\.yaml',
            'busco': r'.*_busco\.txt',
        },
        'path_transformer': {
            r'.*\.fna': '{genome}.{suffix}',
            r'.*\.faa': '{genome}.{suffix}',
            r'.*\.gbk': '{genome}.{suffix}',
            r'.*\.gff': '{genome}.{suffix}',
            r'.*\.sqn': '{genome}.{suffix}',
            r'.*\.ffn': '{genome}.{suffix}',
            r'.*\.emapper.annotations': '{genome}.eggnog',
            r'.*\.[A-Z]{2}': '{genome}.{suffix}',
            r'genome\.md': 'genome.md',
            r'organism\.md': '../../organism.md',
            r'genome\.json': None,  # do not copy this file
            r'organism\.json': None,  # do not copy this file
            r'.*': 'rest/{original_path}',
        }
    }

    def __init__(self, settings: Union[dict, str] = None):
        if settings is None:
            settings = os.environ.get('OGB_IMPORT_SETTINGS', None)

        if type(settings) is dict:
            pass
        elif type(settings) is str:
            with open(settings) as f:
                settings = json.load(f)
        else:
            settings = {}

        self.settings = self.default_settings | settings  # PEP-584: dict union: overwrite defaults with new settings
        self.settings['manual_paths'] = self.default_settings['manual_paths'] | settings.get('manual_paths', {})

        assert set(self.settings.keys()) == set(self.default_settings.keys()), \
            f'OGB_IMPORT_SETTINGS must contain these JSON keys: {set(self.default_settings.keys())}! ' \
            f'reality: {self.settings.keys()}'

        self.organism_template = self.settings['organism_template']
        self.genome_template = self.settings['genome_template']
        self.path_transformer: {re.Pattern: str} = {re.compile(pattern=p): n for p, n in
                                                    self.settings['path_transformer'].items()}

        self.file_finder: {str: re.Pattern} = {n: re.compile(pattern=p) for n, p in
                                               self.settings['manual_paths'].items()}

    def get_path(self, original_path: str, genome: str, organism: str) -> Optional[str]:
        pattern: re.Pattern
        for pattern, new_path in self.path_transformer.items():
            if pattern.fullmatch(string=original_path) is not None:
                if new_path is None:
                    return None  # do not copy this file
                return new_path.format(
                    original_path=original_path,
                    suffix=original_path.rsplit('.', 1)[-1],
                    genome=genome,
                    organism=organism,
                    assembly=genome.rsplit('.', 1)[0]
                )
        raise AssertionError(
            f'File/folder {original_path} does not match any regex specified in import_settings: {self.settings}')

    def find_files(self, files, type: str, root_dir: str) -> [str]:
        pattern = self.file_finder[type]
        matches = [f for f in files if pattern.fullmatch(string=f) is not None]
        if not matches:
            _old_wd = os.getcwd()
            os.chdir(root_dir)
            matches = glob(pattern.pattern)
            os.chdir(_old_wd)
        return matches


class OgbImporter:
    genome_json: dict = None
    organism_json: dict = None

    def __init__(self, folder_structure_dir: str, import_dir: str, organism: str = None, genome: str = None,
                 import_settings: str = None):
        """
        Easily import files into the OpenGenomeBrowser folder structure.

        :param import_dir: Folder with files to import. Required: [.fna, .faa, .gbk, .gff] Optional: [.ffn, .sqn, custom-annotation-files]
        :param folder_structure_dir: Path to the root of the OpenGenomeBrowser folder structure. (Must contain 'organisms' folder.)
        :param organism: Name of the organism.
        :param genome: Identifier of the genome. Must start with organism. May be identical to organism.
        :param rename: Locus tag prefixes must match the genome identifier. If this is not the case, this script can automatically rename relevant files.
        :param import_settings: Path to import settings file. Alternatively, set the environment variable OGB_IMPORT_SETTINGS.
        """
        assert folder_structure_dir is not None and os.path.isdir(
            folder_structure_dir), f'Cannot import PGAP files: {folder_structure_dir=} does not exist.'
        assert os.path.isdir(import_dir), f'Cannot import PGAP files: {import_dir=} does not exist.'

        self.import_settings = ImportSettings(import_settings)

        self.tempdir = tempfile.TemporaryDirectory()

        self.import_dir = import_dir
        self.folder_structure_dir = folder_structure_dir
        self.organisms_dir = f'{folder_structure_dir}/organisms'
        assert os.path.isdir(
            self.organisms_dir), f'{folder_structure_dir=} does not point to a directory that contains an organisms-folder!'

        files = os.listdir(import_dir)
        self.fna = self.find_file(files, 'fna', FastaFile)  # assembly
        self.gbk = self.find_file(files, 'gbk', GenBankFile)  # genbank
        self.ffn = self.find_or_create_ffn(files=files)  # nucleic acid sequences
        self.faa = self.find_or_create_faa(files=files)  # protein
        self.gff = self.find_file(files, 'gff', GffFile)  # general feature format
        self.sqn = self.find_file(files, 'sqn', GenomeFile, raise_error=False)  # GenBank submission file
        self.custom_annotations = self.find_custom_annotations(files)  # custom annotation files / eggnog files

        self.genome_md = self.get_genome_file(files, 'genome.md', GenomeFile, raise_error=False)
        self.organism_md = self.get_genome_file(files, 'organism.md', GenomeFile, raise_error=False)

        self.rest_files = files

        auto_organism, auto_genome = self.detect_organism_genome()
        self.organism = organism if organism else auto_organism
        self.genome = genome if genome else auto_genome
        assert self.genome.startswith(
            self.organism), f'The genome identifier must start with the organism name! {self.genome=} {self.organism=}'

        self.target_dir = f'{self.organisms_dir}/{self.organism}/genomes/{self.genome}'

        assert not os.path.isdir(
            self.target_dir), f'Could not import {self.organism}:{self.genome}: {self.target_dir=} already exists!'

    def __repr__(self):
        return f'{self.organism}::{self.genome}'

    def _append_new_path_to_file(self, file: GenomeFile) -> Optional[str]:
        if file is None:
            return None
        new_path = self.get_new_path(original_path=file.original_path)
        file.new_path = new_path
        file.target_path = os.path.join(self.target_dir, file.new_path)
        return new_path

    def get_new_path(self, original_path: str) -> str:
        return self.import_settings.get_path(original_path=original_path, genome=self.genome, organism=self.organism)

    def add_files_to_json(self, genome_json: dict) -> dict:
        genome_json['cds_tool_faa_file'] = self._append_new_path_to_file(self.faa)
        genome_json['cds_tool_ffn_file'] = self._append_new_path_to_file(self.ffn)
        genome_json['cds_tool_gbk_file'] = self._append_new_path_to_file(self.gbk)
        genome_json['cds_tool_gff_file'] = self._append_new_path_to_file(self.gff)
        genome_json['cds_tool_sqn_file'] = self._append_new_path_to_file(self.sqn)
        genome_json['assembly_fasta_file'] = self._append_new_path_to_file(self.fna)
        genome_json['custom_annotations'] = [
            {'date': ca.date_str(), 'file': self._append_new_path_to_file(ca), 'type': ca.custom_annotation_type}
            for ca in self.custom_annotations
        ]
        return genome_json

    def load_yaml_metadata(self) -> (dict, dict):
        organism_yaml, genome_yaml = {}, {}
        try:
            submol_yaml = self.find_file(files=self.rest_files, key='yaml', file_class=GenomeFile, remove=False)
        except FileNotFoundError:
            return organism_yaml, genome_yaml
        with open(submol_yaml.path) as f:
            submol_yaml = yaml.safe_load(f)

        if 'organism' in submol_yaml and 'genus_species' in submol_yaml['organism']:
            organism_yaml['taxid'] = entrez_organism_to_taxid(submol_yaml['organism']['genus_species'])

        if 'biosample' in submol_yaml:
            genome_yaml['biosample_accession'] = submol_yaml['biosample']
        if 'bioproject' in submol_yaml:
            genome_yaml['bioproject_accession'] = submol_yaml['bioproject']

        if 'publications' in submol_yaml and len(submol_yaml['publications']):
            genome_yaml['literature_references'] = [{
                'url': f"https://pubmed.ncbi.nlm.nih.gov/{p['publication']['pmid']}/",
                'name': shorten(p['publication']['title'], width=30)}
                for p in submol_yaml['publications']
            ]

        return organism_yaml, genome_yaml

    def load_busco_metadata(self) -> dict:
        try:
            busco_file = self.find_file(files=self.rest_files, key='busco', file_class=GenomeFile, remove=False)
        except FileNotFoundError:
            return {}

        return dict(BUSCO=parse_busco(busco_file.path))

    def load_cog_metadata(self) -> dict:
        for file in self.custom_annotations:
            if type(file) is EggnogFile:
                try:
                    cog = file.cog_categories()
                    return {'COG': cog}
                except AssertionError as e:
                    logging.info(f'Failed to extract COG information from {file.path}. {str(e)}')
                    pass
        return {}  # not eggnog file

    def gather_metadata(self):
        '''
        Load metadata from:
          - pgap_submol.yaml
          - *.gbk
          - *_busco.txt
          - organism.json and genome.json
        :return:
        '''

        # start with dummy jsons
        organism_json = organism_json_dummy.copy()
        genome_json = genome_json_dummy.copy()

        # add import_settings
        organism_json.update(self.import_settings.organism_template)
        genome_json.update(self.import_settings.genome_template)

        # add pgap_submol.yaml
        organism_yaml, genome_yaml = self.load_yaml_metadata()
        organism_json.update(organism_yaml)
        genome_json.update(genome_yaml)

        # add *.gbk
        organism_gbk, genome_gbk = self.gbk.metadata()
        organism_json.update(organism_gbk)
        genome_json.update(genome_gbk)

        # add _busco.txt
        genome_json.update(self.load_busco_metadata())

        # add COG from eggnog
        genome_json.update(self.load_cog_metadata())

        # add organism.json from folder structure
        organism_json = merge_json(organism_json, os.path.join(self.target_dir, '../../organism.json'))

        # add organism.json / genome.json from import_dir
        organism_json = merge_json(organism_json,
                                   self.get_file(files=self.rest_files, file='organism.json', raise_error=False))
        genome_json = merge_json(genome_json,
                                 self.get_file(files=self.rest_files, file='genome.json', raise_error=False))

        # add elementary identifiers
        organism_json['name'] = self.organism
        organism_json['representative'] = self.genome
        genome_json['identifier'] = self.genome

        # add files
        genome_json = self.add_files_to_json(genome_json)

        # validate metadata files
        try:
            organism_json_schema.validate(organism_json)
        except SchemaError as e:
            logging.warning(f'{self}: FAILED TO CREATE A VALID organism.json! {str(e)}')
            raise e

        try:
            genome_json_schema.validate(genome_json)
        except SchemaError as e:
            logging.warning(f'{self}: FAILED TO CREATE A VALID genome.json! {str(e)}')
            raise e

        self.organism_json = organism_json
        self.genome_json = genome_json

    def detect_organism_genome(self) -> (str, str):
        strain, locus_tag_prefix = self.gbk.detect_strain_locus_tag_prefix()
        organism, genome = strain, locus_tag_prefix.rstrip('_')
        logging.info(f'autodetected from gbk: {organism=} {genome=}')
        return organism, genome

    def find_or_create_ffn(self, files: [str]) -> GenomeFile:
        try:
            return self.find_file(files, 'ffn', FastaFile)
        except FileNotFoundError:
            self.ffn = self.gbk.path[:-4] + '.ffn'
            logging.info('Creating .ffn based on .gbk...')
            self.gbk.create_ffn(ffn=self.ffn)
            return self.find_file(os.listdir(self.import_dir), 'ffn', FastaFile)

    def find_or_create_faa(self, files: [str]) -> GenomeFile:
        try:
            return self.find_file(files, 'faa', FastaFile)
        except FileNotFoundError:
            self.faa = self.gbk.path[:-4] + '.faa'
            logging.info('Creating .faa based on .gbk...')
            self.gbk.create_faa(faa=self.faa)
            return self.find_file(os.listdir(self.import_dir), 'faa', FastaFile)

    def check_files(self) -> None:
        locus_tag_prefix = f'{self.genome}_'
        self.gbk.validate_locus_tags(locus_tag_prefix=locus_tag_prefix)
        self.gff.validate_locus_tags(locus_tag_prefix=locus_tag_prefix)
        self.faa.validate_locus_tags(locus_tag_prefix=locus_tag_prefix)
        self.ffn.validate_locus_tags(locus_tag_prefix=locus_tag_prefix)
        for ca in self.custom_annotations:
            ca.validate_locus_tags(locus_tag_prefix=locus_tag_prefix)

    def _get_temp(self, file: str) -> str:
        return os.path.join(self.tempdir.name, os.path.basename(file))

    def rename_all(self, new_locus_tag_prefix: str, old_locus_tag_prefix: str = None):
        if not old_locus_tag_prefix:
            old_locus_tag_prefix = self.gbk.detect_locus_tag_prefix()
        assert new_locus_tag_prefix != old_locus_tag_prefix, \
            f'old and new locus_tag_prefix are the same! {old_locus_tag_prefix}, {new_locus_tag_prefix=}'

        kwargs = dict(new_locus_tag_prefix=new_locus_tag_prefix, old_locus_tag_prefix=old_locus_tag_prefix)

        self.gbk.rename(out=self._get_temp(self.gbk.path), **kwargs)
        self.gff.rename(out=self._get_temp(self.gff.path), **kwargs)
        self.faa.rename(out=self._get_temp(self.faa.path), **kwargs)
        self.ffn.rename(out=self._get_temp(self.ffn.path), **kwargs)
        for custom_annotation in self.custom_annotations:
            custom_annotation.rename(out=self._get_temp(custom_annotation.path), **kwargs)

    def perform_import(self):
        assert self.genome_json is not None and self.organism_json is not None, \
            f'Cannot perform import yet. Metadata jsons are missing. ' \
            f'(Run gather_metadata or set OgbImporter.genome_json and OgbImporter.organism_json manually)'

        def copy(src: str, dst: str):
            if os.path.exists(dst):
                logging.warning(f'Overwriting: {src} -> {dst}')
            copy_fn = shutil.copy2 if os.path.isfile(src) else shutil.copytree
            os.makedirs(os.path.dirname(dst), exist_ok=True)  # create parent dir if nonexistent
            copy_fn(src=src, dst=dst)

        os.makedirs(self.target_dir)

        for file in [self.fna, self.faa, self.gbk, self.gff, self.sqn, self.ffn, *self.custom_annotations]:
            if file:
                logging.info(f'{file.original_path} >>copy key file>> {file.new_path}')
                copy(src=file.path, dst=file.target_path)

        for original_path in self.rest_files:
            path = os.path.join(self.import_dir, original_path)
            new_path = self.get_new_path(original_path=original_path)
            if new_path is None or new_path == '':
                logging.info(f'not copying: {original_path}')
                continue
            target_path = os.path.join(self.target_dir, new_path)
            logging.info(f'{original_path} >>copy rest file>> {new_path}')

            copy(src=path, dst=target_path)

        with open(os.path.join(self.target_dir, 'genome.json'), 'w') as f:
            json.dump(self.genome_json, f, indent=4)

        with open(os.path.join(self.target_dir, '../../organism.json'), 'w') as f:
            json.dump(self.organism_json, f, indent=4)

    def get_file(self, files: [str], file: str,
                 raise_error: bool = True, alternative=None, rel_path=False, remove: bool = False) -> str:
        if file in files:
            return file if rel_path else os.path.join(self.import_dir, file)
        elif raise_error:
            raise FileNotFoundError(f'Could not find {file}! {files=}')
        else:
            return alternative

    def get_genome_file(self, files: [str], file: str, file_class: type,
                        raise_error: bool = True, alternative=None, remove: bool = False) -> GenomeFile:
        try:
            file = self.get_file(files=files, file=file, raise_error=True, rel_path=True, remove=remove)
        except FileNotFoundError as e:
            if raise_error:
                raise e
            else:
                return alternative
        genome_file = file_class(os.path.join(self.import_dir, file), original_path=file)
        if remove:
            files.remove(file)
        return genome_file

    def find_file(self, files: [str], key: str, file_class: type,
                  raise_error: bool = True, alternative=None, remove: bool = True
                  ) -> Union[GenomeFile, FastaFile, GffFile, GenBankFile, EggnogFile, CustomAnnotationFile]:
        matches = self.import_settings.find_files(files, type=key, root_dir=self.import_dir)
        if len(matches) == 0:
            if raise_error:
                raise FileNotFoundError(f'Could not find a {key} file! {files=}')
            else:
                return alternative
        if len(matches) > 1:
            raise FileExistsError(f'Found multiple {key} files! {matches=}')
        file = matches[0]
        if remove and file in files:
            files.remove(file)

        genome_file = file_class(os.path.join(self.import_dir, file), original_path=file)
        return genome_file

    def find_custom_annotations(self, files: [str]) -> [
        Union[GenomeFile, FastaFile, GffFile, GenBankFile, EggnogFile, CustomAnnotationFile]]:
        custom_annotations = [self.get_genome_file(files, f, CustomAnnotationFile, remove=True)
                              for f in files
                              if re.fullmatch(pattern=r'.*\.[A-Z]{2}', string=f) is not None]
        eggnog = self.find_file(files, 'eggnog', EggnogFile, raise_error=False)  # GenBank submission file
        if eggnog:
            eggnog: EggnogFile
            logging.info(f'Found eggnog file: {eggnog}')
            custom_annotations.append(eggnog)
        return custom_annotations

    def __del__(self):
        if hasattr(self, 'tempdir'):
            logging.info(f'Deleting {self.tempdir.name}')
            try:
                self.tempdir.cleanup()
            except Exception as e:
                logging.info(f'Could not delete {self.tempdir.name} {str(e)=}')


def import_genome(
        import_dir: str,
        folder_structure_dir: str = None,
        organism: str = None,
        genome: str = None,
        rename: bool = False,
        check_files: bool = True,
        import_settings: str = None
):
    """
    Easily import files into OpenGenomeBrowser folder structure.

    :param import_dir: Folder with files to import. Required: [.fna, .faa, .gbk, .gff] Optional: [.ffn, .sqn, custom-annotation-files]
    :param folder_structure_dir: Path to the root of the OpenGenomeBrowser folder structure. (Must contain 'organisms' folder.)
    :param organism: Name of the organism.
    :param genome: Identifier of the genome. Must start with organism. May be identical to organism.
    :param rename: Locus tag prefixes must match the genome identifier. If this is not the case, this script can automatically rename relevant files.
    :param check_files: If true, check if locus tag prefixes match genome identifier.
    :param import_settings: Path to import settings file. Alternatively, set the environment variable OGB_IMPORT_SETTINGS.
    """
    if folder_structure_dir is None:
        assert 'FOLDER_STRUCTURE' in os.environ, f'Cannot find the folder_structure. Please set --folder_structure_dir or environment variable FOLDER_STRUCTURE'
        folder_structure_dir = os.environ['FOLDER_STRUCTURE']

    current_folder_structure_version = get_folder_structure_version(folder_structure_dir)
    assert current_folder_structure_version == __folder_structure_version__, \
        f'Before importing any genomes, the folder structure needs to be updated to match OpenGenomeBrowser Tools.\n' \
        f'Current version: {current_folder_structure_version}, expected: {__folder_structure_version__}\n' \
        f'Use the script update_folder_structure perform the upgrade!'

    ogb_importer = OgbImporter(folder_structure_dir=folder_structure_dir, import_dir=import_dir, organism=organism,
                               genome=genome, import_settings=import_settings)

    if rename:
        ogb_importer.rename_all(new_locus_tag_prefix=f'{ogb_importer.genome}_')

    ogb_importer.gather_metadata()

    if check_files:
        ogb_importer.check_files()

    ogb_importer.perform_import()


def main():
    import fire

    fire.Fire(import_genome)


if __name__ == '__main__':
    main()



from .utils import GenomeFile, split_locus_tag


class CustomAnnotationFile(GenomeFile):
    def __init__(self, file: str, original_path: str = None, custom_annotation_type: str = None):
        if custom_annotation_type:
            self.custom_annotation_type = custom_annotation_type
        else:
            self.custom_annotation_type = file.rsplit('.', 1)[-1]
        super().__init__(file=file, original_path=original_path)

    def rename(self, out: str, new_locus_tag_prefix: str, old_locus_tag_prefix: str = None, validate: bool = False) -> None:
        old_locus_tag_prefix = self._pre_rename_check(out, new_locus_tag_prefix, old_locus_tag_prefix)

        with open(self.path) as in_f:
            content = in_f.readlines()

        def rename_line(line: str):
            assert line.startswith(old_locus_tag_prefix), f'custom_annotations_file line does not contain old_locus_tag_prefix!' \
                                                          f'{old_locus_tag_prefix=}, {line=}, {self.path=}'
            return line.replace(old_locus_tag_prefix, new_locus_tag_prefix, 1)

        content = [rename_line(line) for line in content]

        with open(out, 'w') as out_f:
            out_f.writelines(content)

        self.path = out

        if validate:
            self.validate_locus_tags(locus_tag_prefix=new_locus_tag_prefix)

    def detect_locus_tag_prefix(self) -> str:
        with open(self.path) as f:
            line = f.readline()
            locus_tag = line.split('\t', 1)[0]
            locus_tag_prefix, gene_id = split_locus_tag(locus_tag)
            assert gene_id.isdigit(), f'locus_tag in {self.path=} is malformed. expected: {locus_tag_prefix}_[0-9]+ reality: {locus_tag}'
            return locus_tag_prefix

    def validate_locus_tags(self, locus_tag_prefix: str = None):
        if locus_tag_prefix is None:
            locus_tag_prefix = self.detect_locus_tag_prefix()

        with open(self.path) as f:
            for line in f:
                locus_tag = line.split('\t')[0]
                real_locus_tag_prefix, gene_id = split_locus_tag(locus_tag)

                assert real_locus_tag_prefix == locus_tag_prefix, \
                    f'locus_tag_prefix in {self.path=} does not match. expected: {locus_tag_prefix} reality: {real_locus_tag_prefix}'
                assert gene_id.isdigit(), f'locus_tag in {self.path=} is malformed. expected: {locus_tag_prefix}_[0-9]+ reality: {locus_tag}'


def rename_custom_annotations(file: str, out: str, new_locus_tag_prefix: str, old_locus_tag_prefix: str = None, validate: bool = False):
    """
    Change the locus tags in a custom annotations file

    :param file: input file
    :param out: output file
    :param new_locus_tag_prefix: desired locus tag
    :param old_locus_tag_prefix: locus tag to replace
    :param validate: if true, perform sanity check
    """

    CustomAnnotationFile(
        file=file
    ).rename(
        out=out,
        new_locus_tag_prefix=new_locus_tag_prefix,
        old_locus_tag_prefix=old_locus_tag_prefix,
        validate=validate
    )


def main():
    import fire

    fire.Fire(rename_custom_annotations)


if __name__ == '__main__':
    main()



import os
import logging
from urllib import request
from tempfile import TemporaryDirectory
from .utils import Entrez, decompress_gz
from .reindex_assembly import reindex_assembly
from .rename_genbank import GenBankFile
from .rename_gff import GffFile
from .genbank_to_fasta import GenBankToFasta


def get_record_id(assembly_name: str) -> str:
    # fetch assembly report
    record = Entrez.read(Entrez.esearch(db='assembly', term=assembly_name, report='full'))
    assert record['Count'] == '1', f'Expected 1 record. Found {record["Count"]}. ' \
                                   f'Ensure a search on https://www.ncbi.nlm.nih.gov/assembly/?term={assembly_name} yields exactlyone result.'

    # extract record id
    ids = record['IdList']
    assert len(ids) == 1, f'Failed to get unique record_id. {ids=}, {assembly_name=}'
    record_id = ids[0]

    logging.info(f'Extracted {record_id=} from {assembly_name=}')
    return record_id


def download_ncbi_file(record_id: str, out: str, file: str):
    assert not os.path.isfile(out), f'Output file already exists! {out=}'

    # fetch summary
    summary = Entrez.read(Entrez.esummary(db='assembly', id=record_id, report='full'))

    # extract ftp link
    url = summary['DocumentSummarySet']['DocumentSummary'][0]['FtpPath_RefSeq']
    assert url.startswith('ftp://ftp.ncbi.nlm.nih.gov/genomes'), f'Url does not match: {url=}'
    assembly_label = os.path.basename(url)
    ftp_link = os.path.join(url, assembly_label + file)
    logging.info(f'Extracted {ftp_link=} from {record_id=}')

    with TemporaryDirectory() as tempdir:
        # download
        gz_file = f'{tempdir}/{assembly_label}.{file}.gz'
        request.urlretrieve(ftp_link, gz_file)
        logging.info(f'Downloaded {ftp_link=}')

        # decompress
        decompress_gz(gz=gz_file, out=out)
        logging.info(f'Decompressed {out=}')


def download_ncbi_fna_gbk_gff(assembly_name: str, out_dir: str) -> (str, str, str):
    record_id = get_record_id(assembly_name)

    fna = os.path.join(out_dir, f'{assembly_name}.fna')
    gbk = os.path.join(out_dir, f'{assembly_name}.gbk')
    gff = os.path.join(out_dir, f'{assembly_name}.gff')

    download_ncbi_file(record_id=record_id, out=fna, file=f'_genomic.fna.gz')
    download_ncbi_file(record_id=record_id, out=gbk, file=f'_genomic.gbff.gz')
    download_ncbi_file(record_id=record_id, out=gff, file=f'_genomic.gff.gz')

    return fna, gbk, gff


def rename_ncbi_files(
        raw_fna: str, raw_gbk: str, raw_gff: str,
        out_fna: str, out_gbk: str, out_gff: str, out_ffn: str, out_faa: str,
        new_locus_tag_prefix: str,
        scaffold_prefix: str,
        old_locus_tag_prefix: str = None,
        leading_zeroes: int = None,
        validate: bool = True
):
    if old_locus_tag_prefix is None:
        old_locus_tag_prefix = GenBankFile(raw_gbk).detect_locus_tag_prefix()

    if not new_locus_tag_prefix.endswith('_'):
        logging.warning(f'new_locus_tag_prefix does not begin with _ (underline): {new_locus_tag_prefix}')

    # reindex assembly
    reindex_assembly(file=raw_fna, out=out_fna, prefix=scaffold_prefix, leading_zeroes=leading_zeroes)

    # rename gff
    gff = GffFile(raw_gff)
    gff.rename(new_locus_tag_prefix=new_locus_tag_prefix, old_locus_tag_prefix=old_locus_tag_prefix, out=out_gff,
               validate=validate)

    # rename gbk
    gbk = GenBankFile(raw_gbk)
    gbk.rename(new_locus_tag_prefix=new_locus_tag_prefix, old_locus_tag_prefix=old_locus_tag_prefix, out=out_gbk,
               validate=validate, scf_prefix=scaffold_prefix, scf_leading_zeroes=leading_zeroes)

    # produce ffn
    GenBankToFasta.convert(gbk=out_gbk, out=out_ffn, format='ffn', strict=validate)

    # produce faa
    GenBankToFasta.convert(gbk=out_gbk, out=out_faa, format='faa', strict=validate)


def download_ncbi_genome(
        assembly_name: str,
        out_dir: str,
        new_locus_tag_prefix: str,
        scaffold_prefix: str = None,
        scaffold_leading_zeroes: int = 1,
        validate: bool = True
):
    """
    Download assembly (fna, gbk, gff), rename the locus tags, and generate CDS FASTAs (faa, ffn)

    :param assembly_name: Name of the assembly on NCBI (e.g. 'GCF_005864195.1')
    :param out_dir: Output directory
    :param new_locus_tag_prefix: The desired locus tag prefix
    :param scaffold_prefix: The desired scaffold prefix of the assembly file (fna)
    :param scaffold_leading_zeroes: format scaffold counter with leading zeroes, e.g.: 5 -> >PREFIX_00001
    :param validate: If True, validate the newly created files
    """

    if not new_locus_tag_prefix.endswith('_'):
        logging.warning(f'new_locus_tag_prefix does not begin with _ (underline): {new_locus_tag_prefix}')

    filename_prefix = new_locus_tag_prefix.rstrip('_')
    out_fna = os.path.join(out_dir, f'{filename_prefix}.fna')
    out_gbk = os.path.join(out_dir, f'{filename_prefix}.gbk')
    out_gff = os.path.join(out_dir, f'{filename_prefix}.gff')
    out_ffn = os.path.join(out_dir, f'{filename_prefix}.ffn')
    out_faa = os.path.join(out_dir, f'{filename_prefix}.faa')

    if scaffold_prefix is None:
        scaffold_prefix = f'{filename_prefix}_scf'

    if not os.path.isdir(out_dir):
        os.makedirs(out_dir, exist_ok=True)

    for file in (out_fna, out_gbk, out_gff, out_ffn, out_faa):
        assert not os.path.isfile(file), f'Output file already exists! {file=}'

    with TemporaryDirectory() as tempdir:
        raw_fna, raw_gbk, raw_gff = download_ncbi_fna_gbk_gff(assembly_name=assembly_name, out_dir=tempdir)

        rename_ncbi_files(
            raw_fna, raw_gbk, raw_gff,
            out_fna, out_gbk, out_gff, out_ffn, out_faa,
            new_locus_tag_prefix=new_locus_tag_prefix,
            scaffold_prefix=scaffold_prefix, leading_zeroes=scaffold_leading_zeroes,
            validate=validate
        )


def main():
    import fire

    fire.Fire(download_ncbi_genome)


if __name__ == '__main__':
    main()



import os
from .folder_looper import FolderLooper


def init_orthofinder(folder_structure_dir: str = None, skip_ignored: bool = True, sanity_check: bool = True, representatives_only: bool = False):
    if folder_structure_dir is None:
        assert 'FOLDER_STRUCTURE' in os.environ, f'Cannot find the folder_structure. Please set --folder_structure_dir or environment variable FOLDER_STRUCTURE'
        folder_structure_dir = os.environ['FOLDER_STRUCTURE']

    orthofinder_dir = os.path.join(folder_structure_dir, 'OrthoFinder')
    fasta_dir = os.path.join(orthofinder_dir, 'fastas')

    if not os.path.isdir(orthofinder_dir):
        os.makedirs(orthofinder_dir)

    assert len(os.listdir(orthofinder_dir)) == 0, f'Error: {orthofinder_dir=} is not empty!'

    os.makedirs(fasta_dir)

    folder_looper = FolderLooper(folder_structure_dir)

    n_faas = 0
    print(f'Linking protein fastas to {fasta_dir}/{{identifier}}.faa')
    for genome in folder_looper.genomes(skip_ignored=skip_ignored, sanity_check=sanity_check, representatives_only=representatives_only):
        faa_path = f"{genome.path}/{genome.get_json_attr('cds_tool_faa_file')}"
        assert os.path.isfile(faa_path), faa_path
        rel_path = os.path.relpath(faa_path, start=fasta_dir)
        os.symlink(src=rel_path, dst=f'{fasta_dir}/{genome.identifier}.faa')
        n_faas += 1

    cmd = f'orthofinder -f {fasta_dir}'
    container_cmd = f'-it --rm -v {folder_structure_dir}:/input:Z davidemms/orthofinder orthofinder -f /input/OrthoFinder/fastas'
    podman_cmd = f'podman run --ulimit=host {container_cmd}'
    docker_cmd = f'docker run --ulimit nofile=1000000:1000000 {container_cmd}'

    print(
        f'Done: Found {n_faas} faas.',
        f'',
        f'Ways to run OrthoFinder:',
        f'1) local install:',
        f'    {cmd}',
        f'',
        f'2) using podman:',
        f'    {podman_cmd}',
        f'',
        f'2) using docker:',
        f'    {docker_cmd}',
        f'',
        f'If you have many genomes, consider run OrthoFinder with the -og option.',
        f'Be sure to to use the -a and -t options!',
        f'More info on https://github.com/davidemms/OrthoFinder',
        sep='\n'
    )


def main():
    import fire

    fire.Fire(init_orthofinder)


if __name__ == '__main__':
    main()



# based on https://realpython.com/pypi-publish-python-package
# How to upload:
#  - change package version in `setup.py` and `__init__.py`
#  - `python setup.py sdist`
#  - `twine upload dist/opengenomebrowser-tools-?.tar.gz`
import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / 'README.md').read_text()

# This call to setup() does all the work
setup(
    name='opengenomebrowser-tools',
    version='0.0.6',
    description='Set of scripts to aid OpenGenomeBrowser administrators import data',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/opengenomebrowser/opengenomebrowser-tools',
    author='Thomas Roder',
    author_email='roder.thomas@gmail.com',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
    ],
    packages=['opengenomebrowser_tools'],
    include_package_data=True,  # see MANIFEST.in
    install_requires=['schema', 'biopython', 'termcolor', 'fire', 'pyyaml'],
    entry_points={
        'console_scripts': [
            'init_folder_structure=opengenomebrowser_tools.init_folder_structure:main',
            'import_genome=opengenomebrowser_tools.import_genome:main',
            'download_ncbi_genome=opengenomebrowser_tools.download_ncbi_genome:main',
            'genbank_to_fasta=opengenomebrowser_tools.genbank_to_fasta:main',
            'reindex_assembly=opengenomebrowser_tools.reindex_assembly:main',
            'rename_custom_annotations=opengenomebrowser_tools.rename_custom_annotations:main',
            'rename_eggnog=opengenomebrowser_tools.rename_eggnog:main',
            'rename_fasta=opengenomebrowser_tools.rename_fasta:main',
            'rename_genbank=opengenomebrowser_tools.rename_genbank:main',
            'rename_gff=opengenomebrowser_tools.rename_gff:main',
            'init_orthofinder=opengenomebrowser_tools.init_orthofinder:main',
            'import_orthofinder=opengenomebrowser_tools.import_orthofinder:main',
            'folder_looper=opengenomebrowser_tools.folder_looper:main',
            'update_folder_structure=opengenomebrowser_tools.update_folder_structure:main',
        ]
    },
)
